; Public Domain file CONCAT tool by Eric Auer 2004.
; Use at your own risk, error checking is limited!
; To assemble ("compile"), use http://nasm.sourceforge.net/
; nasm -o concat.com concat.asm

	org 0x100		; for a .com program

start:	mov ax,sp		; get some memory hint
	dec ax			; if it was 0
	cmp ax,buffer+0x8100	; at least 32k + 256 bytes for buffer?
	mov al,0xff		; errorlevel 255 out of memory
	jb exit

	xor bh,bh
	mov bl,[ds:0x80]	; length of command line
	mov byte [bx+0x81],0	; add string terminator
	mov dx,0x82		; first argument, assuming 1 space
	; between command and argument (use "command/argument" to enforce)
	mov ax,0x3d00		; open for read, returns handle AX
	int 0x21		; DOS
	jc exit
	jmp short okay

close1:	mov al,0xfe		; errorlevel 254, target open error
	jmp short close		; (target is NOT created, it must EXIST!)

close2:	push ax
	mov bx,[ds:target]	; file handle
	mov ah,0x3e		; close, handle BX
	int 0x21		; DOS
	pop ax
close:	push ax
	mov bx,[ds:handle]	; file handle
	mov ah,0x3e		; close, handle BX
	int 0x21		; DOS
	pop ax
exit:	cmp al,0		; any nonzero errorlevel?
	jz fine			; else do not show help
	push ax
	mov dx,help		; help message
	mov ah,9		; show string
	int 0x21		; DOS
	pop ax
fine:	mov ah,0x4c		; exit, errorlevel AL
	int 0x21

okay:	mov [handle],ax		; store file handle
	
	mov dx,target		; target file
	mov ax,0x3d01		; open for write only
	int 0x21		; DOS
	jc close1
	mov [target],ax		; store file handle

	mov bx,ax		; file handle
	xor cx,cx		; high 16 bits of offset: 0
	xor dx,dx		; low 16 bits of offset: 0
	mov ax,0x4202		; seek to position relative to end of file
	int 0x21		; DOS
	jc close2

	; DX:AX is now current absolute offset in file, and as we went
	; to "end plus 0" it is also equal to the file size ... if
	; anybody wants to know the filesize at that moment ;-).

	; you can even seek to "end PLUS something" (usually you do MINUS)
	; ... which means "append garbage (or zeroes?) to make the file
	; have such an offset" - but only if you write to the file later.

copy_loop:

	mov bx,[ds:handle]	; file handle
	mov dx,buffer		; buffer position
	mov cx,0x8000		; try reading 32k
	mov ah,0x3f		; read file, returns "AX bytes read"
	int 0x21		; DOS
	jc close2
	or ax,ax		; nothing read anymore?
	jz close2		; end of input file, nothing read
				; returns errorlevel 0, okay

	mov bx,[ds:target]	; file handle
	mov cx,ax		; try writing N bytes after reading N...
	mov dx,buffer		; buffer position
	mov ah,0x40		; write to file, returns "AX bytes written"
				; (writing 0 bytes just sets file size)
	int 0x21		; DOS
	jc close2
	cmp ax,cx		; less bytes written than intended?
	mov al,0xfd		; errorlevel 253, disk full / write error
	jnz close2		; if write gave problems
	jmp short copy_loop	; else keep copying data!

	; Note: target file will end exactly at the place where the
	; last write ends. So if that is not at the end of the file,
	; you first have to seek to the end and write 0 extra bytes,
	; or the file will be truncated. No problem for us here.

help	db "Free CONCAT by Eric Auer 2004, appends a file to target.bin.",13,10
	db "Usage: CONCAT File",13,10
	db "File must not be same as target, or you might get into a loop!",13,10

	; In redirected DOSEmu drive, files are always updated at once --> LOOP
	; (you read file, write contents, read what you just wrote...)
	; Result: If target.bin was non-empty, it will grow to "infinite" size.

	; In normal FAT drive, file size / EOF position is NOT updated until
	; you close the file: You read file, write contents, but run out of
	; data to read at the position where the file ended before you started
	; writing to it. So the file will simply become twice the size.

	db "Errorlevels: 0 okay, 254 no writeable target.bin found",13,10
	db "255 out of memory, 253 disk full, 1..127: source read error."
	db 13,10,"$"

handle	dw 0			; source file handle
target	db "target.bin",0	; file to which input file is appended
				; overwritten by target file handle

	align 16
buffer:				; label - buffer starts here
